@@ -7,8 +7,10 @@ use DynaLoader ();
 
 use vars qw($VERSION @ISA);
 use vars qw{$err $errstr $drh $sqlite_version};
+use vars qw{%COLLATION};
+
 BEGIN {
-    $VERSION = '1.26_01';
+    $VERSION = '1.26_03';
     @ISA     = 'DynaLoader';
 
     # Initialize errors
@@ -24,13 +26,41 @@ BEGIN {
 
 __PACKAGE__->bootstrap($VERSION);
 
+tie %COLLATION, 'DBD::SQLite::_WriteOnceHash';
+$COLLATION{perl}       = sub { $_[0] cmp $_[1] };
+$COLLATION{perllocale} = sub { use locale; $_[0] cmp $_[1] };
+
+my $methods_are_installed;
+
 sub driver {
-    $drh or
+    return $drh if $drh;
+
+    if (!$methods_are_installed && $DBI::VERSION >= 1.608) {
+        DBI->setup_driver('DBD::SQLite');
+
+        DBD::SQLite::db->install_method('sqlite_last_insert_rowid');
+        DBD::SQLite::db->install_method('sqlite_busy_timeout');
+        DBD::SQLite::db->install_method('sqlite_create_function');
+        DBD::SQLite::db->install_method('sqlite_create_aggregate');
+        DBD::SQLite::db->install_method('sqlite_create_collation');
+        DBD::SQLite::db->install_method('sqlite_collation_needed');
+        DBD::SQLite::db->install_method('sqlite_progress_handler');
+        DBD::SQLite::db->install_method('sqlite_commit_hook');
+        DBD::SQLite::db->install_method('sqlite_rollback_hook');
+        DBD::SQLite::db->install_method('sqlite_update_hook');
+        DBD::SQLite::db->install_method('sqlite_set_authorizer');
+        DBD::SQLite::db->install_method('sqlite_backup_from_file');
+        DBD::SQLite::db->install_method('sqlite_backup_to_file');
+        DBD::SQLite::db->install_method('sqlite_enable_load_extension');
+        $methods_are_installed++;
+    }
+
     $drh = DBI::_new_drh( "$_[0]::dr", {
         Name        => 'SQLite',
         Version     => $VERSION,
         Attribution => 'DBD::SQLite by Matt Sergeant et al',
     } );
+    return $drh;
 }
 
 sub CLONE {
@@ -65,7 +95,7 @@ sub connect {
 
     # To avoid unicode and long file name problems on Windows,
     # convert to the shortname if the file (or parent directory) exists.
-    if ( $^O =~ /MSWin32/ and $real ne ':memory:' ) {
+    if ( $^O =~ /MSWin32/ and $real ne ':memory:' and $real ne '') {
         require Win32;
         require File::Basename;
         my ($file, $dir, $suffix) = File::Basename::fileparse($real);
@@ -86,11 +116,15 @@ sub connect {
     # Hand off to the actual login function
     DBD::SQLite::db::_login($dbh, $real, $user, $auth) or return undef;
 
-    # Install perl collations
-    my $perl_collation        = sub { $_[0] cmp $_[1] };
-    my $perl_locale_collation = sub { use locale; $_[0] cmp $_[1] };
-    $dbh->func( "perl",       $perl_collation,        "create_collation" );
-    $dbh->func( "perllocale", $perl_locale_collation, "create_collation" );
+    # Register the on-demand collation installer
+    $DBI::VERSION >= 1.608 
+      ? $dbh->sqlite_collation_needed(\&install_collation)
+      : $dbh->func(\&install_collation, "collation_needed");
+
+    # Register the REGEXP function 
+    $DBI::VERSION >= 1.608 
+      ? $dbh->sqlite_create_function("REGEXP", 2, \&regexp)
+      : $dbh->func("REGEXP", 2, \&regexp, "create_function");
 
     # HACK: Since PrintWarn = 0 doesn't seem to actually prevent warnings
     # in DBD::SQLite we set Warn to false if PrintWarn is false.
@@ -101,6 +135,25 @@ sub connect {
     return $dbh;
 }
 
+
+sub install_collation {
+  my ($dbh, $collation_name) = @_;
+  my $collation = $DBD::SQLite::COLLATION{$collation_name}
+    or die "can't install, unknown collation : $collation_name";
+  $DBI::VERSION >= 1.608 
+      ? $dbh->sqlite_create_collation($collation_name => $collation)
+      : $dbh->func($collation_name => $collation, "create_collation");
+}
+
+# default implementation for sqlite 'REGEXP' infix operator.
+# Note : args are reversed, i.e. "a REGEXP b" calls REGEXP(b, a)
+# (see http://www.sqlite.org/vtab.html#xfindfunction)
+sub regexp { 
+  use locale;
+  return scalar($_[1] =~ $_[0]);
+}
+
+
 package DBD::SQLite::db;
 
 sub prepare {
@@ -133,16 +186,28 @@ sub get_info {
     return $v;
 }
 
+sub _attached_database_list {
+    my $dbh = shift;
+    my @attached;
+
+    my $sth_databases = $dbh->prepare( 'PRAGMA database_list' );
+    $sth_databases->execute;
+    while ( my $db_info = $sth_databases->fetchrow_hashref ) {
+        push @attached, $db_info->{name} if $db_info->{seq} >= 2;
+    }
+    return @attached;
+}
+
 # SQL/CLI (ISO/IEC JTC 1/SC 32 N 0595), 6.63 Tables
 # Based on DBD::Oracle's
-# See also http://www.ch-werner.de/sqliteodbc/html/sqliteodbc_8c.html#a117
+# See also http://www.ch-werner.de/sqliteodbc/html/sqlite3odbc_8c.html#a213
 sub table_info {
     my ($dbh, $cat_val, $sch_val, $tbl_val, $typ_val) = @_;
 
     my @where = ();
     my $sql;
-    if (   defined($cat_val) && $cat_val eq '%'
-       && defined($sch_val) && $sch_val eq '' 
+    if (  defined($cat_val) && $cat_val eq '%'
+       && defined($sch_val) && $sch_val eq ''
        && defined($tbl_val) && $tbl_val eq '')  { # Rule 19a
             $sql = <<'END_SQL';
 SELECT NULL TABLE_CAT
@@ -152,21 +217,28 @@ SELECT NULL TABLE_CAT
      , NULL REMARKS
 END_SQL
     }
-    elsif (   defined($sch_val) && $sch_val eq '%' 
-          && defined($cat_val) && $cat_val eq '' 
+    elsif (  defined($cat_val) && $cat_val eq ''
+          && defined($sch_val) && $sch_val eq '%'
           && defined($tbl_val) && $tbl_val eq '') { # Rule 19b
             $sql = <<'END_SQL';
 SELECT NULL      TABLE_CAT
-     , NULL      TABLE_SCHEM
+     , t.tn      TABLE_SCHEM
      , NULL      TABLE_NAME
      , NULL      TABLE_TYPE
      , NULL      REMARKS
+FROM (
+     SELECT 'main' tn
+     UNION SELECT 'temp' tn
 END_SQL
+            for my $db_name (_attached_database_list($dbh)) {
+                $sql .= "     UNION SELECT '$db_name' tn\n";
+            }
+            $sql .= ") t\n";
     }
-    elsif (    defined($typ_val) && $typ_val eq '%' 
-           && defined($cat_val) && $cat_val eq '' 
-           && defined($sch_val) && $sch_val eq '' 
-           && defined($tbl_val) && $tbl_val eq '') { # Rule 19c
+    elsif (  defined($cat_val) && $cat_val eq ''
+          && defined($sch_val) && $sch_val eq ''
+          && defined($tbl_val) && $tbl_val eq ''
+          && defined($typ_val) && $typ_val eq '%') { # Rule 19c
             $sql = <<'END_SQL';
 SELECT NULL TABLE_CAT
      , NULL TABLE_SCHEM
@@ -187,26 +259,41 @@ SELECT *
 FROM
 (
 SELECT NULL         TABLE_CAT
-     , NULL         TABLE_SCHEM
+     ,              TABLE_SCHEM
      , tbl_name     TABLE_NAME
      ,              TABLE_TYPE
      , NULL         REMARKS
      , sql          sqlite_sql
 FROM (
-    SELECT tbl_name, upper(type) TABLE_TYPE, sql
+    SELECT 'main' TABLE_SCHEM, tbl_name, upper(type) TABLE_TYPE, sql
     FROM sqlite_master
     WHERE type IN ( 'table','view')
 UNION ALL
-    SELECT tbl_name, 'LOCAL TEMPORARY' TABLE_TYPE, sql
+    SELECT 'temp' TABLE_SCHEM, tbl_name, 'LOCAL TEMPORARY' TABLE_TYPE, sql
     FROM sqlite_temp_master
     WHERE type IN ( 'table','view')
+END_SQL
+
+            for my $db_name (_attached_database_list($dbh)) {
+                    $sql .= <<"END_SQL";
 UNION ALL
-    SELECT 'sqlite_master'      tbl_name, 'SYSTEM TABLE' TABLE_TYPE, NULL sql
+    SELECT '$db_name' TABLE_SCHEM, tbl_name, upper(type) TABLE_TYPE, sql
+    FROM "$db_name".sqlite_master
+    WHERE type IN ('table','view')
+END_SQL
+            }
+
+            $sql .= <<'END_SQL';
 UNION ALL
-    SELECT 'sqlite_temp_master' tbl_name, 'SYSTEM TABLE' TABLE_TYPE, NULL sql
+    SELECT 'main' TABLE_SCHEM, 'sqlite_master'      tbl_name, 'SYSTEM TABLE' TABLE_TYPE, NULL sql
+UNION ALL
+    SELECT 'temp' TABLE_SCHEM, 'sqlite_temp_master' tbl_name, 'SYSTEM TABLE' TABLE_TYPE, NULL sql
 )
 )
 END_SQL
+            if ( defined $sch_val ) {
+                    push @where, "TABLE_SCHEM LIKE '$sch_val'";
+            }
             if ( defined $tbl_val ) {
                     push @where, "TABLE_NAME LIKE '$tbl_val'";
             }
@@ -401,8 +488,34 @@ sub column_info {
     return $sth;
 }
 
+
+#======================================================================
+# An internal tied hash package used for %DBD::SQLite::COLLATION, to
+# prevent people from unintentionally overriding globally registered collations.
+
+package DBD::SQLite::_WriteOnceHash;
+require Tie::Hash;
+
+our @ISA = qw(Tie::StdHash);
+
+sub TIEHASH {
+  bless {}, $_[0];
+}
+
+sub STORE {
+  ! exists $_[0]->{$_[1]} or die "entry $_[1] already registered";
+  $_[0]->{$_[1]} = $_[2];
+}
+
+sub DELETE {
+  die "deletion of entry $_[1] is forbidden";
+}
+
 1;
 
+
+
+
 __END__
 
 =pod
@@ -414,15 +527,15 @@ DBD::SQLite - Self-contained RDBMS in a DBI Driver
 =head1 SYNOPSIS
 
   use DBI;
-  my $dbh = DBI->connect("dbi:SQLite:dbname=dbfile","","");
+  my $dbh = DBI->connect("dbi:SQLite:dbname=$dbfile","","");
 
 =head1 DESCRIPTION
 
-SQLite is a public domain RDBMS database engine that you can find
-at L<http://www.sqlite.org/>.
+SQLite is a public domain file-based relational database engine that
+you can find at L<http://www.sqlite.org/>.
 
-Rather than ask you to install SQLite first, because SQLite is public
-domain, B<DBD::SQLite> includes the entire thing in the distribution.
+B<DBD::SQLite> is a Perl DBI driver for SQLite, that includes 
+the entire thing in the distribution.
 So in order to get a fast transaction capable RDBMS working for your
 perl project you simply have to install this module, and B<nothing>
 else.
@@ -463,6 +576,29 @@ details about core features.
 Currently many statement attributes are not implemented or are
 limited by the typeless nature of the SQLite database.
 
+=head1 CONNECTING
+
+The name of the database file is passed in the the DBI connection 
+string :
+
+  my $dbh = DBI->connect("dbi:SQLite:dbname=$dbfile","","");
+
+The file is opened in read/write mode, and will be created if
+it does not exist yet. 
+
+If the filename C<$dbfile> is ":memory:", then a private, temporary
+in-memory database is created for the connection. This in-memory
+database will vanish when the database connection is closed. Future
+versions of SQLite might make use of additional special filenames that
+begin with the ":" character. It is recommended that when a database
+filename actually does begin with a ":" character you should prefix
+the filename with a pathname such as "./" to avoid ambiguity.
+
+If the filename C<$dbfile> is an empty string, then a private,
+temporary on-disk database will be created. This private database will
+be automatically deleted as soon as the database connection is closed.
+
+
 =head1 DRIVER PRIVATE ATTRIBUTES
 
 =head2 Database Handle Attributes
@@ -476,8 +612,9 @@ e.g., "2.8.0". Can only be read.
 
 =item unicode
 
-If set to a true value, B<DBD::SQLite> will turn the UTF-8 flag on for all text
-strings coming out of the database (this feature is currently disabled for perl < 5.8.5). For more details on the UTF-8 flag see
+If set to a true value, B<DBD::SQLite> will turn the UTF-8 flag on for all
+text strings coming out of the database (this feature is currently disabled
+for perl < 5.8.5). For more details on the UTF-8 flag see
 L<perlunicode>. The default is for the UTF-8 flag to be turned off.
 
 Also note that due to some bizarreness in SQLite's type system (see
@@ -500,7 +637,16 @@ Defining the column type as C<BLOB> in the DDL is B<not> sufficient.
 
 =head1 DRIVER PRIVATE METHODS
 
-=head2 $dbh->func('last_insert_rowid')
+The following methods can be called via the func() method with a little
+tweak, but the use of func() method is now discouraged by the L<DBI> author
+for various reasons (see DBI's document
+L<http://search.cpan.org/dist/DBI/lib/DBI/DBD.pm#Using_install_method()_to_expose_driver-private_methods>
+for details). So, if you're using L<DBI> >= 1.608, use these C<sqlite_>
+methods. If you need to use an older L<DBI>, you can call these like this:
+
+  $dbh->func( ..., "(method name without sqlite_ prefix)" );
+
+=head2 $dbh->sqlite_last_insert_rowid()
 
 This method returns the last inserted rowid. If you specify an INTEGER PRIMARY
 KEY as the first column in your table, that is the column that is returned.
@@ -512,17 +658,17 @@ method instead. The usage of this is:
   $h->last_insert_id($catalog, $schema, $table_name, $field_name [, \%attr ])
 
 Running C<$h-E<gt>last_insert_id("","","","")> is the equivalent of running
-C<$dbh-E<gt>func('last_insert_rowid')> directly.
+C<$dbh-E<gt>sqlite_last_insert_rowid()> directly.
 
-=head2 $dbh->func('busy_timeout')
+=head2 $dbh->sqlite_busy_timeout()
 
 Retrieve the current busy timeout.
 
-=head2 $dbh->func( $ms, 'busy_timeout' )
+=head2 $dbh->sqlite_busy_timeout( $ms )
 
 Set the current busy timeout. The timeout is in milliseconds.
 
-=head2 $dbh->func( $name, $argc, $code_ref, "create_function" )
+=head2 $dbh->sqlite_create_function( $name, $argc, $code_ref )
 
 This method will register a new function which will be useable in an SQL
 query. The method's parameters are:
@@ -548,63 +694,74 @@ This should be a reference to the function's implementation.
 For example, here is how to define a now() function which returns the
 current number of seconds since the epoch:
 
-  $dbh->func( 'now', 0, sub { return time }, 'create_function' );
+  $dbh->sqlite_create_function( 'now', 0, sub { return time } );
 
 After this, it could be use from SQL as:
 
   INSERT INTO mytable ( now() );
 
-=head2 $dbh->func( $name, $code_ref, "create_collation" )
 
-This method will register a new function which will be useable in an SQL
-query as a COLLATE option for sorting. The method's parameters are:
+=head3 REGEXP function
+
+SQLite includes syntactic support for an infix operator 'REGEXP', but
+without any implementation. The C<DBD::SQLite> driver
+automatically registers an implementation that performs standard
+perl regular expression matching, using current locale. So for example
+you can search for words starting with an 'A' with a query like
+
+  SELECT * from table WHERE column REGEXP '\bA\w+'
+
+If you want case-insensitive searching, use perl regex flags, like this :
+
+  SELECT * from table WHERE column REGEXP '(?i:\bA\w+)'
+
+The default REGEXP implementation can be overriden through the
+C<create_function> API described above.
+
+Note that regexp matching will B<not> use SQLite indices, but will iterate
+over all rows, so it could be quite costly in terms of performance.
+
+
+=head2 $dbh->sqlite_create_collation( $name, $code_ref )
+
+This method manually registers a new function which will be useable in an SQL
+query as a COLLATE option for sorting. Such functions can also be registered
+automatically on demand: see section L</"COLLATION FUNCTIONS"> below.
+
+The method's parameters are:
 
 =over
 
 =item $name
 
-The name of the function. This is the name of the function as it will
-be used from SQL.
+The name of the function exposed to SQL.
 
 =item $code_ref
 
-This should be a reference to the function's implementation.
+Reference to the function's implementation.
 The driver will check that this is a proper sorting function.
 
 =back
 
-Collations C<binary> and C<nocase> are builtin within SQLite.
-Collations C<perl> and C<perllocale> are builtin within 
-the B<DBD::SQLite> driver, and correspond to the 
-Perl C<cmp> operator with or without the L<locale> pragma; 
-so you can write for example 
 
-  CREATE TABLE foo(
-      txt1 COLLATE perl,
-      txt2 COLLATE perllocale,
-      txt3 COLLATE nocase
-  )
+=head2 $dbh->sqlite_collation_needed( $code_ref )
 
-or
+This method manually registers a callback function that will
+be invoked whenever an undefined collation sequence is required
+from an SQL statement. The callback is invoked as 
 
-  SELECT * FROM foo ORDER BY name COLLATE perllocale
+  $code_ref->($dbh, $collation_name)
 
-If the attribute C<< $dbh->{unicode} >> is set, strings coming from
-the database and passed to the collation function will be properly
-tagged with the utf8 flag; but this only works if the 
-C<unicode> attribute is set B<before> the call to 
-C<create_collation>. The recommended way to activate unicode
-is to set the parameter at connection time :
+and should register the desired collation using 
+L</"sqlite_create_collation">. 
 
-  my $dbh = DBI->connect(
-      "dbi:SQLite:dbname=foo", "", "", 
-      {
-          RaiseError => 1,
-          unicode    => 1,
-      }
-  );
+An initial callback is already registered by C<DBD::SQLite>,
+so for most common cases it will be simpler to just
+add your collation sequences in the C<%DBD::SQLite::COLLATION>
+hash (see section L</"COLLATION FUNCTIONS"> below).
 
-=head2 $dbh->func( $name, $argc, $pkg, 'create_aggregate' )
+
+=head2 $dbh->sqlite_create_aggregate( $name, $argc, $pkg )
 
 This method will register a new aggregate function which can then be used
 from SQL. The method's parameters are:
@@ -688,7 +845,7 @@ Here is a simple aggregate function which returns the variance
       return $sigma;
   }
   
-  $dbh->func( "variance", 1, 'variance', "create_aggregate" );
+  $dbh->sqlite_create_aggregate( "variance", 1, 'variance' );
 
 The aggregate function can then be used as:
 
@@ -698,7 +855,7 @@ The aggregate function can then be used as:
 
 For more examples, see the L<DBD::SQLite::Cookbook>.
 
-=head2 $dbh->func( $n_opcodes, $code_ref, 'progress_handler' )
+=head2 $dbh->sqlite_progress_handler( $n_opcodes, $code_ref )
 
 This method registers a handler to be invoked periodically during long
 running calls to SQLite.
@@ -713,7 +870,7 @@ large query. The parameters are:
 The progress handler is invoked once for every C<$n_opcodes>
 virtual machine opcodes in SQLite.
 
-=item $handler
+=item $code_ref
 
 Reference to the handler subroutine.  If the progress handler returns
 non-zero, the SQLite operation is interrupted. This feature can be used to
@@ -724,6 +881,342 @@ progress handler.
 
 =back
 
+=head2 $dbh->sqlite_commit_hook( $code_ref )
+
+This method registers a callback function to be invoked whenever a
+transaction is committed. Any callback set by a previous call to
+C<sqlite_commit_hook> is overridden. A reference to the previous
+callback (if any) is returned.  Registering an C<undef> disables the
+callback.
+
+When the commit hook callback returns zero, the commit operation is
+allowed to continue normally. If the callback returns non-zero, then
+the commit is converted into a rollback (in that case, any attempt to
+I<explicitly> call C<< $dbh->rollback() >> afterwards would yield an
+error).
+
+=head2 $dbh->sqlite_rollback_hook( $code_ref )
+
+This method registers a callback function to be invoked whenever a
+transaction is rolled back. Any callback set by a previous call to
+C<sqlite_rollback_hook> is overridden. A reference to the previous
+callback (if any) is returned.  Registering an C<undef> disables the
+callback.
+
+
+=head2 $dbh->sqlite_update_hook( $code_ref )
+
+This method registers a callback function to be invoked whenever a row
+is updated, inserted or deleted. Any callback set by a previous call to
+C<sqlite_update_hook> is overridden. A reference to the previous
+callback (if any) is returned.  Registering an C<undef> disables the
+callback.
+
+The callback will be called as
+
+  $code_ref->($action_code, $database, $table, $rowid)
+
+where
+
+=over
+
+=item $action_code
+
+is an integer equal to either C<DBD::SQLite::INSERT>,
+C<DBD::SQLite::DELETE> or C<DBD::SQLite::UPDATE>
+(see L</"Action Codes">);
+
+=item $database
+
+is the name of the database containing the affected row;
+
+=item $table
+
+is the name of the table containing the affected row;
+
+=item $rowid
+
+is the unique 64-bit signed integer key of the affected row within that table.
+
+=back
+
+
+=head2 $dbh->sqlite_set_authorizer( $code_ref )
+
+This method registers an authorizer callback to be invoked whenever
+SQL statements are being compiled by the L<DBI/prepare> method.  The
+authorizer callback should return C<DBD::SQLite::OK> to allow the
+action, C<DBD::SQLite::IGNORE> to disallow the specific action but
+allow the SQL statement to continue to be compiled, or
+C<DBD::SQLite::DENY> to cause the entire SQL statement to be rejected
+with an error. If the authorizer callback returns any other value,
+then then C<prepare> call that triggered the authorizer will fail with
+an error message.
+
+An authorizer is used when preparing SQL statements from an untrusted
+source, to ensure that the SQL statements do not try to access data
+they are not allowed to see, or that they do not try to execute
+malicious statements that damage the database. For example, an
+application may allow a user to enter arbitrary SQL queries for
+evaluation by a database. But the application does not want the user
+to be able to make arbitrary changes to the database. An authorizer
+could then be put in place while the user-entered SQL is being
+prepared that disallows everything except SELECT statements.
+
+The callback will be called as
+
+  $code_ref->($action_code, $string1, $string2, $database, $trigger_or_view)
+
+where
+
+=over
+
+=item $action_code
+
+is an integer that specifies what action is being authorized
+(see L</"Action Codes">).
+
+=item $string1, $string2
+
+are strings that depend on the action code
+(see L</"Action Codes">).
+
+=item $database
+
+is the name of the database (C<main>, C<temp>, etc.) if applicable.
+
+=item $trigger_or_view
+
+is the name of the inner-most trigger or view that is responsible for
+the access attempt, or C<undef> if this access attempt is directly from
+top-level SQL code.
+
+=back
+
+
+
+=head2 $dbh->sqlite_backup_from_file( $filename )
+
+This method accesses the SQLite Online Backup API, and will take a backup of
+the named database file, copying it to, and overwriting, your current database
+connection. This can be particularly handy if your current connection is to the
+special :memory: database, and you wish to populate it from an existing DB.
+
+=head2 $dbh->sqlite_backup_to_file( $filename )
+
+This method accesses the SQLite Online Backup API, and will take a backup of
+the currently connected database, and write it out to the named file.
+
+=head2 $dbh->sqlite_enable_load_extension( $bool )
+
+Calling this method with a true value enables loading (external)
+sqlite3 extensions. After the call, you can load extensions like this:
+
+  $dbh->sqlite_enable_load_extension(1);
+  $sth = $dbh->prepare("select load_extension('libsqlitefunctions.so')")
+  or die "Cannot prepare: " . $dbh->errstr();
+
+
+=head1 DRIVER CONSTANTS
+
+A subset of SQLite C constants are made available to Perl,
+because they may be needed when writing
+hooks or authorizer callbacks. For accessing such constants, 
+the C<DBD::Sqlite> module must be explicitly C<use>d at compile
+time. For example, an authorizer that forbids any
+DELETE operation would be written as follows :
+
+  use DBD::SQLite;
+  $dbh->sqlite_set_authorizer(sub {
+    my $action_code = shift;
+    return $action_code == DBD::SQLite::DELETE ? DBD::SQLite::DENY
+                                               : DBD::SQLite::OK;
+  });
+
+The list of constants implemented in C<DBD::SQLite> is given 
+below; more information can be found ad 
+at L<http://www.sqlite.org/c3ref/constlist.html>.
+
+=head2 Authorizer Return Codes
+
+  OK
+  DENY
+  IGNORE
+
+=head2 Action Codes
+
+The L</set_authorizer> method registers a callback function that is
+invoked to authorize certain SQL statement actions. The first
+parameter to the callback is an integer code that specifies what
+action is being authorized. The second and third parameters to the 
+callback are strings, the meaning of which varies according to the 
+action code. Below is the list of action codes, together with their
+associated strings.
+
+  # constant              string1         string2
+  # ========              =======         =======
+  CREATE_INDEX            Index Name      Table Name
+  CREATE_TABLE            Table Name      undef
+  CREATE_TEMP_INDEX       Index Name      Table Name
+  CREATE_TEMP_TABLE       Table Name      undef
+  CREATE_TEMP_TRIGGER     Trigger Name    Table Name
+  CREATE_TEMP_VIEW        View Name       undef
+  CREATE_TRIGGER          Trigger Name    Table Name
+  CREATE_VIEW             View Name       undef
+  DELETE                  Table Name      undef
+  DROP_INDEX              Index Name      Table Name
+  DROP_TABLE              Table Name      undef
+  DROP_TEMP_INDEX         Index Name      Table Name
+  DROP_TEMP_TABLE         Table Name      undef
+  DROP_TEMP_TRIGGER       Trigger Name    Table Name
+  DROP_TEMP_VIEW          View Name       undef
+  DROP_TRIGGER            Trigger Name    Table Name
+  DROP_VIEW               View Name       undef
+  INSERT                  Table Name      undef
+  PRAGMA                  Pragma Name     1st arg or undef
+  READ                    Table Name      Column Name
+  SELECT                  undef           undef
+  TRANSACTION             Operation       undef
+  UPDATE                  Table Name      Column Name
+  ATTACH                  Filename        undef
+  DETACH                  Database Name   undef
+  ALTER_TABLE             Database Name   Table Name
+  REINDEX                 Index Name      undef
+  ANALYZE                 Table Name      undef
+  CREATE_VTABLE           Table Name      Module Name
+  DROP_VTABLE             Table Name      Module Name
+  FUNCTION                undef           Function Name
+  SAVEPOINT               Operation       Savepoint Name
+
+
+=head1 COLLATION FUNCTIONS
+
+=head2 Definition
+
+SQLite v3 provides the ability for users to supply arbitrary
+comparison functions, known as user-defined "collation sequences" or
+"collating functions", to be used for comparing two text values.
+L<http://www.sqlite.org/datatype3.html#collation>
+explains how collations are used in various SQL expressions.
+
+=head2 Builtin collation sequences
+
+The following collation sequences are builtin within SQLite :
+
+=over
+
+=item B<BINARY>
+
+Compares string data using memcmp(), regardless of text encoding.
+
+=item B<NOCASE>
+
+The same as binary, except the 26 upper case characters of ASCII are
+folded to their lower case equivalents before the comparison is
+performed. Note that only ASCII characters are case folded. SQLite
+does not attempt to do full UTF case folding due to the size of the
+tables required.
+
+=item B<RTRIM>
+
+The same as binary, except that trailing space characters are ignored.
+
+=back
+
+In addition, C<DBD::SQLite> automatically installs the 
+following collation sequences :
+
+=over
+
+=item B<perl>
+
+corresponds to the Perl C<cmp> operator
+
+=item B<perllocale>
+
+Perl C<cmp> operator, in a context where C<use locale> is activated.
+
+=back
+
+=head2 Usage
+
+You can write for example
+
+  CREATE TABLE foo(
+      txt1 COLLATE perl,
+      txt2 COLLATE perllocale,
+      txt3 COLLATE nocase
+  )
+
+or
+
+  SELECT * FROM foo ORDER BY name COLLATE perllocale
+
+=head2 Unicode handling
+
+If the attribute C<< $dbh->{unicode} >> is set, strings coming from
+the database and passed to the collation function will be properly
+tagged with the utf8 flag; but this only works if the
+C<unicode> attribute is set B<before> the first call to
+a perl collation sequence . The recommended way to activate unicode
+is to set the parameter at connection time :
+
+  my $dbh = DBI->connect(
+      "dbi:SQLite:dbname=foo", "", "",
+      {
+          RaiseError => 1,
+          unicode    => 1,
+      }
+  );
+
+
+=head2 Adding user-defined collations
+
+The native SQLite API for adding user-defined collations is 
+exposed through methods L</"sqlite_create_collation"> and
+L</"sqlite_collation_needed">.
+
+To avoid calling these functions every time a C<$dbh> handle is
+created, C<DBD::SQLite> offers a simpler interface through the
+C<%DBD::SQLite::COLLATION> hash : just insert your own
+collation functions in that hash, and whenever an unknown
+collation name is encountered in SQL, the appropriate collation
+function will be loaded on demand from the hash. For example,
+here is a way to sort text values regardless of their accented
+characters :
+
+  use DBD::SQLite;
+  $DBD::SQLite::COLLATION{no_accents} = sub {
+    my ( $a, $b ) = map lc, @_;
+    tr[]
+      [aaaaaacdeeeeiiiinoooooouuuuy] for $a, $b;
+    $a cmp $b;
+  };
+  my $dbh  = DBI->connect("dbi:SQLite:dbname=dbfile");
+  my $sql  = "SELECT ... FROM ... ORDER BY ... COLLATE no_accents");
+  my $rows = $dbh->selectall_arrayref($sql);
+
+
+The builtin C<perl> or C<perllocale> collations are predefined
+in that same hash.
+
+The COLLATION hash is a global registry within the current process;
+ hence there is a risk of undesired side-effects. Therefore, to
+ prevent action at distance, the hash is implemented as a "write-only"
+ hash, that will happily accept new entries, but will raise an
+ exception if any attempt is made to override or delete a existing
+ entry (including the builtin C<perl> and C<perllocale>).  
+
+If you really, really need to change or delete an entry, you can
+ always grab the tied object underneath C<%DBD::SQLite::COLLATION> ---
+ but don't do that unless you really know what you are doing. Also
+observe that changes in the global hash will not modify existing
+collations in existing database handles: it will only affect new
+I<requests> for collations. In other words, if you want to change
+the behaviour of a collation within an existing C<$dbh>, you 
+need to call the L</create_collation> method directly.
+
+
 =head1 BLOBS
 
 As of version 1.11, blobs should "just work" in SQLite as text columns.
@@ -756,7 +1249,7 @@ database file must be writable by SQLite because the library will create
 several temporary files there.
 
 To access the database from the command line, try using dbish which comes with
-the DBI module. Just type:
+the L<DBI::Shell> module. Just type:
 
   dbish dbi:SQLite:foo.db
 
@@ -767,14 +1260,17 @@ with B<DBD::SQLite> and use the supplied C<sqlite> command line tool.
 
 =head1 FUNCTIONS AND BIND PARAMETERS
 
-As of this writing, a SQL that compares a return value of a function with a numeric bind value like this doesn't work as you might expect.
+As of this writing, a SQL that compares a return value of a function with a
+numeric bind value like this doesn't work as you might expect.
 
   my $sth = $dbh->prepare(q{
     SELECT bar FROM foo GROUP BY bar HAVING count(*) > ?;
   });
   $sth->execute(5);
 
-This is because DBD::SQLite assumes that all the bind values are text (and should be quoted) by default. Thus the above statement becomes like this while executing:
+This is because DBD::SQLite assumes that all the bind values are text (and
+should be quoted) by default. Thus the above statement becomes like this
+while executing:
 
   SELECT bar FROM foo GROUP BY bar HAVING count(*) > "5";
 
@@ -784,10 +1280,11 @@ There are two workarounds for this.
 
 =item Use bind_param() explicitly
 
-As shown above in the C<BLOB> section, you can always use C<bind_param()> to tell the type of a bind value.
+As shown above in the C<BLOB> section, you can always use C<bind_param()> to
+tell the type of a bind value.
 
   use DBI qw(:sql_types);  # Don't forget this
-
+  
   my $sth = $dbh->prepare(q{
     SELECT bar FROM foo GROUP BY bar HAVING count(*) > ?;
   });
@@ -860,6 +1357,15 @@ Implement one or more leak detection tests that only run during
 AUTOMATED_TESTING and RELEASE_TESTING and validate that none of the C
 code we work with leaks.
 
+=head2 Stream API for Blobs
+
+Reading/writing into blobs using C<sqlite2_blob_open> / C<sqlite2_blob_close>.
+
+=head2 Flags for sqlite3_open_v2
+
+Support the full API of sqlite3_open_v2 (flags for opening the file).
+
+
 =head1 SUPPORT
 
 Bugs should be reported via the CPAN bug tracker at
@@ -896,6 +1402,8 @@ Adam Kennedy E<lt>adamk@cpan.orgE<gt>
 
 Max Maischein E<lt>corion@cpan.orgE<gt>
 
+Laurent Dami E<lt>dami@cpan.orgE<gt>
+
 =head1 COPYRIGHT
 
 The bundled SQLite code in this distribution is Public Domain.
